/*

    File: Copy (2) of Main.cpp
    Copyright (C) 2000-2007 Christopher Moore (christopher.e.moore@gmail.com)
	  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License along
    with this program; if not, write the Free Software Foundation, Inc., 51
    Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/
#include <windows.h>
#include <gl/gl.h>
#include <math.h>
#include <assert.h>

#include "../m_foundation/main.h"
#include "../m_foundation/const.h"
#include "../m_foundation/random.h"
#include "../m_math3d/quat.h"
#include "../m_math3d/vector.h"
#include "../m_math3d/matrix.h"

/*

typedef float mat4_t[16];
mat4_t mat4_ident = {1,0,0,0,0,1,0,0,0,0,1,0,0,0,0,1};

typedef struct fr_branch_s fr_branch_t;
typedef struct de_branch_s de_branch_t;
typedef struct fr_stack_s fr_stack_t;

struct de_branch_s {
	vector_t	scale;
	vector_t	translate;
	vector_t	rotate;
	int			dest;	//fr_branch_t dest index, post transform
} de_branch[] = {
	{{1,1,1},{0,0,0},{0,0,0},0},
};

struct fr_branch_s {
	int		*branch;		//array of all new de_branch_t which this guy generates
} *fr_branch[] = {
	{0},
};

#define FR_STACK_MAX	32
struct fr_stack_s {
	mat4_t matrix;
	int branch;				//points to fr_branch structure, which points to de_branch to find what to make next
	int child;				//which child of this branch we are currenly processing.  set this in a stack inst before travelling past it.
} fr_stack[FR_STACK_MAX];

// ************** vertexes ************** 

#define	VERTEX_MAX	65536

vector_t vertex_array[VERTEX_MAX];
int vertex_index = 0;

static int last_ring_start = 0;
static int last_ring_end = 0;

void Begin_Ring(void) {
	last_ring_start = vertex_index;
}

void Add_Vertex(float x, float y, float z) {
	if (vertex_index < VERTEX_MAX) {
		vertex_array[vertex_index].x = x;
		vertex_array[vertex_index].y = y;
		vertex_array[vertex_index].z = z;
		vertex_index++;
	}
}

void End_Ring(void) {
	last_ring_end = vertex_index;
}

// ************** fractal ************** 

bool Init_Fractal(void) {
	int stack_depth = 0;
	memcpy(&fr_stack[0].matrix, &mat4_ident, sizeof(mat4_t));
	fr_stack[stack_depth].branch = 0;	//which fr_branch we are processing
	fr_stack[stack_depth].child = 0;	//which of its children we are processing

	//create a ring stack / initialize it *HERE* with a default ring radius/pointcount
	Begin_Ring();
	const int ring_count = 4;
	for (int i = 0; i < ring_count; i++) {	//ring count
		double theta = 2.0 * M_PI * (double)i / (double)ring_count;
		Add_Vertex(cos(theta), sin(theta), 0);
	}
	End_Ring();

	do {
		//test bounds
		if (stack_depth >= FR_STACK_MAX-1) {	//we hit the upper bound - go back!
			stack_depth--;
			if (stack_depth < 0) break;
			thisBranch = &fr_branch[stack_depth];
			fr_stack[stack_depth].child++;
			continue;	//do a child check next iteration
		}

		int newBranch = fr_stack[stack_depth].branch;
		if (newBranch < 0 || newBranch >= numberof(fr_branch)) {
			printf("newBranch out of bounds\n");
			break;
		}

		if (fr_stack[stack_depth].child > fr_branch[newBranch].child_count]) {
			stack_depth--;
			if (stack_depth < 0) break;

		}

		//process from stack_depth to stack_depth+1

		stack_depth++;
		fr_stack[stack_depth];



	//for fr_branch[fr_stack[i].branch]
	//  for all sub-branches
	//  create a child matrix from the parent matrix
	//  generate rings for the child matrix
	//	create polygons between the prev ring and this child ring

	fr_stack[0].branch

	return true;
}

bool Init_Application3D(void) {

	CVar_Exec("consolefont ../m_foundation/consoledata/font.bmp");

	if (!Init_Fractal()) return false;
	
	return true;
}

void Update_Application3D(void) {
}

void Shutdown_Application3D(void) {
}
*/

#define numberof(x) (sizeof(x)/sizeof(x[0]))


class TriBuffer {
	int tri[65536][3];
	int count;
public:
	TriBuffer() {
		memset(tri, 0, sizeof(tri));
		count = 0;
	}

	void addTri(int a, int b, int c) {
		if (count >= numberof(tri)) {
			main_console->Print("tri buffer full\n");
			return;
		}
		tri[count][0] = a;
		tri[count][1] = b;
		tri[count][2] = c;
		count++;
	}

	int getTriVtxIndex(int tri, int corner) {
		return this->tri[tri][corner];
	}

	int getCount() { return count; }
};

typedef struct {
	vector_t x;
	vector_t n;
	float st[2];
} vertex_t;

class VertexBuffer {
private:
	vertex_t vertex[65536];
	int count;

public:

	VertexBuffer() {
		memset(vertex, 0, sizeof(vertex));
		count = 0;
	}

	//returns the first vertex of a set of vertices
	int alloc(int size) {
		if ((int)numberof(vertex) - count < size) {
			main_console->Print("no more room left!\n");
			return 0;
		}
		count += size;
		return count - size;
	}

	void setVtx(int index, vector_t v) {
		if (index < 0 || index >= count) return;
		memset(&vertex[index], 0, sizeof(vertex_t));
		vertex[index].x = v;
	}

	void setTexCoord(int index, float s, float t) {
		if (index < 0 || index >= count) return;
		vertex[index].st[0] = s;
		vertex[index].st[1] = t;
	}

	const float *getVertexPtr(int index) {
		return vertex[index].x.v;
	}

	const float *getNormalPtr(int index) {
		return vertex[index].n.v;
	}

	const float *getTexCoordPtr(int index ) {
		return vertex[index].st;
	}

	int getCount() { return count; }

	void genNormals(TriBuffer *tri) {
		assert(tri);

		for (int i = 0; i < count; i++) {
			vertex[i].n = vectorZero;
		}

		for (i = 0; i < tri->getCount(); i++) {

			vector_t n;

			VectorPlaneUnitNormal(
				&vertex[tri->getTriVtxIndex(i,2)].x,
				&vertex[tri->getTriVtxIndex(i,1)].x,
				&vertex[tri->getTriVtxIndex(i,0)].x, &n);

			VectorAdd(&vertex[tri->getTriVtxIndex(i,0)].n, &n, &vertex[tri->getTriVtxIndex(i,0)].n);
			VectorAdd(&vertex[tri->getTriVtxIndex(i,1)].n, &n, &vertex[tri->getTriVtxIndex(i,1)].n);
			VectorAdd(&vertex[tri->getTriVtxIndex(i,2)].n, &n, &vertex[tri->getTriVtxIndex(i,2)].n);
		}

		for (i = 0; i < count; i++) {
			VectorUnit(&vertex[i].n, &vertex[i].n);
		}

	}
};

typedef struct {
	quat_t		r;	//rotation
	scalar_t	s;	//scale
	vector_t	t;	//translation
} t3d_t;	//3D transform

static t3d_t t3dIdentity;

static t3d_t t3dStack[256];
static int t3dStackIndex = 0;

void T3DStackInit(void) {
	t3dStackIndex = 0;
	t3dStack[t3dStackIndex] = t3dIdentity;
}

bool T3DTestStackBounds(void) {
	if (t3dStackIndex < 0) {
		main_console->Print("T3D Stack Underflow!!!\n");
		return false;
	}
	if (t3dStackIndex >= numberof(t3dStack)) {
		main_console->Print("T3D Stack Overflow!!!\n");
		return false;
	}
	return true;
}

void T3DPush(void) {
	t3dStackIndex++;
	if (!T3DTestStackBounds()) return;
	t3dStack[t3dStackIndex] = t3dStack[t3dStackIndex-1];
}

void T3DPop(void) {
	t3dStackIndex--;
	T3DTestStackBounds();
}

t3d_t *T3DGetCurrent(void) {
	int index = T3DTestStackBounds() ? t3dStackIndex : numberof(t3dStack) - 1;
	return &t3dStack[index];
}

void T3DMul(const t3d_t *a, const t3d_t *b, t3d_t *res) {
	t3d_t c;
	//calc new translation
	VectorScale(&b->t, a->s, &c.t);
	QuatRotateVector(&c.t, &a->r, &c.t);
	VectorAdd(&c.t, &a->t, &c.t);
	//calc new scale
	c.s = a->s * b->s;
	//calc new rotation
	QuatMul(&a->r, &b->r, &c.r);
	//copy into result
	*res = c;
}

void T3DScale(float s) {
	t3d_t *t = T3DGetCurrent();
	t3d_t r = t3dIdentity;
	r.s = s;
	T3DMul(t, &r, t);
}

void T3DTranslate(float x, float y, float z) {
	t3d_t *t = T3DGetCurrent();
	t3d_t r = t3dIdentity;
	r.t.x = x;
	r.t.y = y;
	r.t.z = z;
	T3DMul(t, &r, t);
}

void T3DRotate(float degrees, float x, float y, float z) {
	t3d_t *t = T3DGetCurrent();
	t3d_t r = t3dIdentity;
	r.r.x = x;
	r.r.y = y;
	r.r.z = z;
	r.r.w = degrees;
	VectorUnit(&r.r.vec, &r.r.vec);
	AngleAxisToQuat(&r.r, &r.r);
	T3DMul(t, &r, t);
}

void T3DVertex(const vector_t *b, vector_t *res) {
	vector_t c;
	t3d_t *a = T3DGetCurrent();
	VectorScale(b, a->s, &c);
	QuatRotateVector(&c, &a->r, &c);
	VectorAdd(&c, &a->t, &c);
	*res = c;
}


//app-spec.
Texture texture;
VertexBuffer vertexBuffer;
TriBuffer triBuffer;

void setRelVtx(int index, float x, float y, float z, float s, float t) {
	if (!T3DTestStackBounds()) return;

	vector_t v = {x,y,z};
	T3DVertex(&v, &v);
	vertexBuffer.setVtx(index, v);
	vertexBuffer.setTexCoord(index, s, t);
}

class VertexRange {
public://private:
	int start;
	int size;

//as much as I hate global variables, thats what this is becoming...
	VertexBuffer *vtx;

public:
	//by default reference the single static variable
	//-- this is going to become our main ctor method
	VertexRange() {
		this->vtx = &vertexBuffer;
		this->start = 0;
		this->size = 0;
	}

	VertexRange(VertexBuffer *vtx) {
		this->vtx = vtx;
		this->start = 0;
		this->size = 0;
	}

	void alloc(int size) {
		this->size = size;
		this->start = vtx->alloc(size);
	}

	void setRelVtx(int index, float x, float y, float z, float s, float t) {
		::setRelVtx(start + index, x, y, z, s, t);
	}

	void bind(VertexRange *other, TriBuffer *tri) {
		
		//write poly between the vertices

		for (int i = 0; i < size; i++) {
			int i2 = i;
			if (i2 >= other->size) i2 = other->size - 1;
			tri->addTri(
				start + i,
				start + ((i+1) % size),
				other->start + i2
			);
		}

		for (i = 0; i < other->size; i++) {
			int i2 = i+1;
			if (i2 > size) i2 = size;
			if (i2 == size) i2 = 0;
			tri->addTri(
				start + i2,
				other->start + ((i+1) % other->size),
				other->start + i
			);
		}
	}
};

// **************** SCRIPT INTERFACE ****************

typedef union vtxrange_u {
	unsigned int l;
	struct {
		unsigned short start;
		unsigned short size;
	};
} vtxrange_t;

//allocs from the main vertexBuffer object
int vertexRangeAlloc(int size) {
	vtxrange_t vr;
	vr.size = size;
	vr.start = vertexBuffer.alloc(size);
	return vr.l;
}

//binds to the main triBuffer objects
void BindRange(int l1, int l2) {
	vtxrange_t *v1 = (vtxrange_t *)(&l1);
	vtxrange_t *v2 = (vtxrange_t *)(&l2);

	//write poly between the vertices

	for (int i = 0; i < v1->size; i++) {
		int i2 = i;
		if (i2 >= v2->size) i2 = v2->size - 1;
		triBuffer.addTri(
			v1->start + i,
			v1->start + ((i+1) % v1->size),
			v2->start + i2
		);
	}

	for (i = 0; i < v2->size; i++) {
		int i2 = i+1;
		if (i2 > v1->size) i2 = v1->size;
		if (i2 == v1->size) i2 = 0;
		triBuffer.addTri(
			v1->start + i2,
			v2->start + ((i+1) % v2->size),
			v2->start + i
		);
	}
}

void setRangeVtx(int range, int index, float x, float y, float z, float s, float t) {
	vtxrange_t *v = (vtxrange_t *)(&range);
	setRelVtx(v->start + index, x, y, z, s, t);
}

// **************** BEGIN MF SCRIPT ****************

//we dont have script structures, so 'range' will be a two-short struct
//the low word will be the start
//the high word will be the size

int tree_branch(int life, float width) {
	int range;
	if (life == 0) {
		range = vertexRangeAlloc(1);
		setRangeVtx(range, 0, 0, 0, 0, 0, 0);
		//done - return the range
	} else {
		range = vertexRangeAlloc(3);
		setRangeVtx(range, 0, 0,			-width, 0,	0,		0.46f * (life&1) + 0.02f);
		setRangeVtx(range, 1, width * 0.1f,	0,		0,	1,		0.46f * (life&1) + 0.02f);
		setRangeVtx(range, 2, 0,			width,	0,	0.5f,	0.46f * (life&1) + 0.02f);
//		setRangeVtx(range, 3, -width * 0.1f,0,		0,	1,		0.46f * (life&1) + 0.02f);

		T3DTranslate(0, 0, 10);
		T3DScale(0.8f);
		T3DRotate(20, 0, 1, 0);

		//build child
		T3DPush();
		int nextRange = tree_branch(life-1, width);
		T3DPop();

		//map between the two
		BindRange(range, nextRange);
	}
	return range;
}

int tree_start(int leaves, int life) {

	//create vertices
	int range = vertexRangeAlloc(1);
	setRangeVtx(range, 0, 0,0,0,0,0);

	if (leaves <= 0) return range;

	//build children
	for (int i = 0; i < leaves; i++) {
		//rot 120 degrees relative to world Z axis
		T3DRotate(360.f / (float)leaves, 0, 0, 1);

		T3DPush();
		//drop pitch 50 degrees relative to world Z axis
		T3DRotate(20.f * (float)life + 60.f, 0, 1, 0);

		//link from 'this' to children
		int nextRange = tree_branch(5, 3.f);
		T3DPop();

		//bind between range and nextRange
		BindRange(range, nextRange);
	}

	if (life > 1) {
		T3DRotate(200, 0, 0, 1);

		T3DPush();
		tree_start(leaves, life-1);
		T3DPop();
	}

	return range;
}

int tree_trunk(int life) {

	T3DTranslate(0, 0, 10);
	T3DScale(0.9f);
	T3DRotate(10, 0 + 0.05f * MCRand(), 0.3f + 0.05f * MCRand(), 1 + 0.05f * MCRand());

	float radius = 3.f;

	int range = vertexRangeAlloc(4);
	setRangeVtx(range, 0, radius,	0,			0, 0, 0.98f - 0.46f * (life&1));
	setRangeVtx(range, 1, 0,		radius,		0, 1, 0.98f - 0.46f * (life&1));
	setRangeVtx(range, 2, -radius,	0,			0, 0, 0.98f - 0.46f * (life&1));
	setRangeVtx(range, 3, 0,		-radius,	0, 1, 0.98f - 0.46f * (life&1));

	if (life > 0) {
		T3DPush();
		int nextRange = tree_trunk(life-1);
		T3DPop();
		BindRange(range, nextRange);
	} else {
		T3DPush();
		int nextRange = tree_start(4, 3);
		T3DPop();
		BindRange(range, nextRange);
	}

	return range;
}

// **************** END MF SCRIPT ****************

bool Init_Application3D(void) {
	CVar_Exec("consolefont ../m_foundation/consoledata/font.bmp");
	texture.Create("tree.bmp");

	t3dIdentity.r = quatIdentity;
	t3dIdentity.s = 1.f;
	t3dIdentity.t = vectorZero;

	T3DStackInit();
	tree_trunk(7);

	vertexBuffer.genNormals(&triBuffer);

	main_console->Print("created %d vtxs, %d tris\n",
		vertexBuffer.getCount(), triBuffer.getCount());

	overrideCameraKeys = 1;

	glEnable(GL_DEPTH_TEST);

	glEnable(GL_CULL_FACE);

	if (found_GL_EXT_separate_specular_color) {
		glLightModelf(GL_LIGHT_MODEL_COLOR_CONTROL_EXT, GL_SEPARATE_SPECULAR_COLOR_EXT);
	}

	const static float fv0001[] = {0,0,0,1};
	const static float fv1111[] = {1,1,1,1};
	const static float materialShininess = 10.f;
	const static float materialSpecular[] = {0.2f, 0.15f, 0.1f, 1};

	glEnable(GL_TEXTURE);
	glEnable(GL_TEXTURE_2D);
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	
	glLightModelf(GL_LIGHT_MODEL_LOCAL_VIEWER, 1.f);
	glLightModelf(GL_LIGHT_MODEL_TWO_SIDE, 0.f);
	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, fv0001);

	glMaterialf(GL_FRONT, GL_SHININESS, materialShininess);
	glMaterialfv(GL_FRONT, GL_AMBIENT, fv1111);
	glMaterialfv(GL_FRONT, GL_DIFFUSE, fv1111);
	glMaterialfv(GL_FRONT, GL_SPECULAR, materialSpecular);
	glMaterialfv(GL_FRONT, GL_EMISSION, fv0001);

	glLightfv(GL_LIGHT0, GL_AMBIENT, fv0001);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, fv1111);
	glLightfv(GL_LIGHT0, GL_SPECULAR, fv1111);

//	const static float light0position[] = {0,0,0,1};
//	glLightfv(GL_LIGHT0, GL_POSITION, light0position);

	return true;
}

void Update_Application3D(void) {

	static quat_t light0position = {30, 0, 30, 1};
	static M_CVar *cv = new M_CVar(&light0position, "lightpos");
	glLightfv(GL_LIGHT0, GL_POSITION, light0position.v);

	texture.Set();

	glBegin(GL_TRIANGLES);
	for (int i = 0; i < triBuffer.getCount(); i++) {
		for (int j = 0; j < 3; j++) {
			int index = triBuffer.getTriVtxIndex(i,j);
			glTexCoord2fv(vertexBuffer.getTexCoordPtr(index));
			glNormal3fv(vertexBuffer.getNormalPtr(index));
			glVertex3fv(vertexBuffer.getVertexPtr(index));
		}
	}
	glEnd();
}

void Shutdown_Application3D(void) {
}
