/*

    File: Copy of Main.cpp
    Copyright (C) 2000-2007 Christopher Moore (christopher.e.moore@gmail.com)
	  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License along
    with this program; if not, write the Free Software Foundation, Inc., 51
    Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/
#include <windows.h>
#include <gl/gl.h>
#include <math.h>
#include <assert.h>

#include "../m_foundation/main.h"
#include "../m_math3d/matrix.h"

/*

typedef float mat4_t[16];
mat4_t mat4_ident = {1,0,0,0,0,1,0,0,0,0,1,0,0,0,0,1};

typedef struct fr_branch_s fr_branch_t;
typedef struct de_branch_s de_branch_t;
typedef struct fr_stack_s fr_stack_t;

struct de_branch_s {
	vector_t	scale;
	vector_t	translate;
	vector_t	rotate;
	int			dest;	//fr_branch_t dest index, post transform
} de_branch[] = {
	{{1,1,1},{0,0,0},{0,0,0},0},
};

struct fr_branch_s {
	int		*branch;		//array of all new de_branch_t which this guy generates
} *fr_branch[] = {
	{0},
};

#define FR_STACK_MAX	32
struct fr_stack_s {
	mat4_t matrix;
	int branch;				//points to fr_branch structure, which points to de_branch to find what to make next
	int child;				//which child of this branch we are currenly processing.  set this in a stack inst before travelling past it.
} fr_stack[FR_STACK_MAX];

// ************** vertexes ************** 

#define	VERTEX_MAX	65536

vector_t vertex_array[VERTEX_MAX];
int vertex_index = 0;

static int last_ring_start = 0;
static int last_ring_end = 0;

void Begin_Ring(void) {
	last_ring_start = vertex_index;
}

void Add_Vertex(float x, float y, float z) {
	if (vertex_index < VERTEX_MAX) {
		vertex_array[vertex_index].x = x;
		vertex_array[vertex_index].y = y;
		vertex_array[vertex_index].z = z;
		vertex_index++;
	}
}

void End_Ring(void) {
	last_ring_end = vertex_index;
}

// ************** fractal ************** 

bool Init_Fractal(void) {
	int stack_depth = 0;
	memcpy(&fr_stack[0].matrix, &mat4_ident, sizeof(mat4_t));
	fr_stack[stack_depth].branch = 0;	//which fr_branch we are processing
	fr_stack[stack_depth].child = 0;	//which of its children we are processing

	//create a ring stack / initialize it *HERE* with a default ring radius/pointcount
	Begin_Ring();
	const int ring_count = 4;
	for (int i = 0; i < ring_count; i++) {	//ring count
		double theta = 2.0 * M_PI * (double)i / (double)ring_count;
		Add_Vertex(cos(theta), sin(theta), 0);
	}
	End_Ring();

	do {
		//test bounds
		if (stack_depth >= FR_STACK_MAX-1) {	//we hit the upper bound - go back!
			stack_depth--;
			if (stack_depth < 0) break;
			thisBranch = &fr_branch[stack_depth];
			fr_stack[stack_depth].child++;
			continue;	//do a child check next iteration
		}

		int newBranch = fr_stack[stack_depth].branch;
		if (newBranch < 0 || newBranch >= numberof(fr_branch)) {
			printf("newBranch out of bounds\n");
			break;
		}

		if (fr_stack[stack_depth].child > fr_branch[newBranch].child_count]) {
			stack_depth--;
			if (stack_depth < 0) break;

		}

		//process from stack_depth to stack_depth+1

		stack_depth++;
		fr_stack[stack_depth];



	//for fr_branch[fr_stack[i].branch]
	//  for all sub-branches
	//  create a child matrix from the parent matrix
	//  generate rings for the child matrix
	//	create polygons between the prev ring and this child ring

	fr_stack[0].branch

	return true;
}

bool Init_Application3D(void) {

	CVar_Exec("consolefont ../m_foundation/consoledata/font.bmp");

	if (!Init_Fractal()) return false;
	
	return true;
}

void Update_Application3D(void) {
}

void Shutdown_Application3D(void) {
}
*/

#define numberof(x) (sizeof(x)/sizeof(x[0]))


class TriBuffer {
	int tri[65536][3];
	int count;
public:
	TriBuffer() {
		memset(tri, 0, sizeof(tri));
		count = 0;
	}

	void addTri(int a, int b, int c) {
		if (count >= numberof(tri)) {
			main_console->Print("tri buffer full\n");
			return;
		}
		tri[count][0] = a;
		tri[count][1] = b;
		tri[count][2] = c;
		count++;
	}

	int getTriVtxIndex(int tri, int corner) {
		return this->tri[tri][corner];
	}

	int getCount() { return count; }
};

class VertexBuffer {
private:
	vector_t vertex[65536];
	int count;

public:

	VertexBuffer() {
		memset(vertex, 0, sizeof(vertex));
		count = 0;
	}

	//returns the first vertex of a set of vertices
	int alloc(int size) {
		if ((int)numberof(vertex) - count < size) {
			main_console->Print("no more room left!\n");
			return 0;
		}
		count += size;
		return count - size;
	}

	void setVtx(int index, vector_t v) {
		if (index < 0 || index >= count) return;
		vertex[index] = v;
	}

	const float *getVtxPtr(int index) {
		return vertex[index].v;
	}

	int getCount() { return count; }
};

class TreeStack {
public:
	//the current space
	matrix_t matrix;
	vector_t trans;
	//how much longer this branch will live.
	int life;
	int detail;
};

class TreeRing {

private:
	int start;
	int size;
	VertexBuffer *buffer;

public:
	TreeRing(VertexBuffer *buffer, TreeStack *stack) {
		this->buffer = buffer;
		this->size = stack->detail;
		this->start = buffer->alloc(this->size);

		main_console->Print("gen ring size %d\n", size);

		for (int i = 0; i < size; i++) {
			float theta = (float)(2.0 * M_PI * (double)i / (double)size);
			vector_t vtx = {0, (float)cos(theta), (float)sin(theta)};
			MatrixMul(&stack->matrix, &vtx, &vtx);
			VectorAdd(&stack->trans, &vtx, &vtx);
			buffer->setVtx(start + i, vtx);
		}
	}

	int getSize() { return size; }

	int getVtxIndex(int i) { return i + start; }
};

class TreeNode {
public:
	int marker;	//for memory only
};

/**
 * Class describing transforms from parent to child TreeStack objects between 
 * node iterations.
 * TODO: SCRIPT THIS ALL!
 */
class TreeTransform {
public:
	//for transforming position, angle, scale, skew:
	matrix_t matrix;	//rotation, scale
	vector_t trans;		//translation
	int life_dec;

	bool transform(TreeStack *src, TreeStack *dest) {
		//l9fe
		dest->life = src->life - life_dec;
		if (dest->life < 0) return false;
		//i simplified 4x4 mat mul into a mat & vec 3x3 & 3x1 mul's & add's (removing some mul zeros)
		//translation
		MatrixMul(&src->matrix, &trans, &dest->trans);
		VectorAdd(&dest->trans, &src->trans, &dest->trans);
		//rotation & scale
		MatrixMul(&src->matrix, &matrix, &dest->matrix);
		//detail drop
		dest->detail = src->detail;
		return true;	//succes
	}
};

class TreeEdge {
public:
	//n[parent], n[child]
	TreeNode *n[2];

	TreeTransform transform;

	TreeEdge(TreeNode *n1, TreeNode *n2, TreeTransform transform) {
		this->n[0] = n1;
		this->n[1] = n2;
		this->transform = transform;
	}
};

class Tree {
private:
	TreeNode *node[256];
	int nodeCount;

	TreeEdge *edge[256];
	int edgeCount;

	VertexBuffer *vertexBuffer;
	TriBuffer *triBuffer;
public:
	TreeStack initStack;

	Tree(VertexBuffer *vertexBuffer, TriBuffer *triBuffer) {
		nodeCount = 0;
		edgeCount = 0;
		this->vertexBuffer = vertexBuffer;
		this->triBuffer = triBuffer;
	}

	~Tree() {
		for (int i = 0; i < nodeCount; i++) {
			delete node[i];
		}
		for (i = 0; i < edgeCount; i++) {
			delete edge[i];
		}
	}

	TreeNode *createNode() {
		if (nodeCount >= numberof(node)) return NULL;
		TreeNode *newNode = new TreeNode();
		node[nodeCount] = newNode;
		nodeCount++;
		return newNode;
	}

	bool createEdge(TreeNode *n1, TreeNode *n2, TreeTransform m) {
		if (edgeCount >= numberof(edge)) return false;
		TreeEdge *newEdge = new TreeEdge(n1,n2,m);
		edge[edgeCount] = newEdge;
		edgeCount++;
		return true;
	}

	TreeNode *getRoot() {
		assert(nodeCount > 0);
		return node[0];
	}

	void bind(TreeRing r1, TreeRing r2) {
		
		//write poly between the vertices

		for (int i = 0; i < r1.getSize(); i++) {
			int i2 = i;
			if (i2 >= r2.getSize()) i2 = r2.getSize() - 1;
			triBuffer->addTri(r1.getVtxIndex(i), r1.getVtxIndex( (i+1)%r1.getSize() ), r2.getVtxIndex(i2));
		}

		for (i = 0; i < r2.getSize(); i++) {
			int i2 = i+1;
			if (i2 > r1.getSize()) i2 = r1.getSize();
			if (i2 == r1.getSize()) i2 = 0;
			triBuffer->addTri(r2.getVtxIndex(i), r2.getVtxIndex( (i+1)%r2.getSize() ), r1.getVtxIndex(i2));
		}
	}

	void process(TreeNode *node, TreeRing ring, TreeStack stack) {
		for (int i = 0; i < edgeCount; i++) {
			TreeEdge *e = edge[i];
			if (e->n[0] != node) continue;
			TreeStack childStack;	//write local vars to the stack frame rather than using parametrs.  negative addressing rather than positive, but it all goes to the same place.

			if (!e->transform.transform(&stack, &childStack))  continue;	//this child is done.
			//left-mul transform the current matrix by the edge matrix to get the child's current matrix

			TreeNode *child = e->n[1];
			TreeRing childRing(vertexBuffer, &childStack);
			bind(ring, childRing);
			process(child, childRing, childStack);
		}
	}

	void build() {
		TreeStack stack = initStack;
		TreeNode *rootNode = getRoot();
		TreeRing rootRing(vertexBuffer, &initStack);

		process(rootNode, rootRing, initStack);
	}
};

//app-spec.
Texture texture;
VertexBuffer vertexBuffer;
TriBuffer triBuffer;

bool Init_Application3D(void) {
	CVar_Exec("consolefont ../m_foundation/consoledata/font.bmp");
	texture.Create("surface.bmp");

	Tree *tree = new Tree(&vertexBuffer, &triBuffer);

	//ctor subclass
	{
		tree->initStack.matrix = matrixIdentity;
		tree->initStack.trans = vectorZero;
		tree->initStack.life = 10;				//five lives like a half cat.
		tree->initStack.detail = 3;

		//define node states
		TreeNode *nA = tree->createNode();
		TreeNode *nB = tree->createNode();

		//construct the tree
		TreeTransform transform;
		//transform.matrix = matrixIdentity;
		MatrixRotation(0.5, 1, 0, 0, &transform.matrix);
		MatrixMul(&transform.matrix, 0.9f, &transform.matrix);
		VectorSet(&transform.trans, 3, 0, 0);
		transform.life_dec = 1;		//decrement life.

		tree->createEdge(nA, nA, transform);
		tree->createEdge(nB, nA, transform);

		//transform.matrix = matrixIdentity;
		MatrixRotation(1, 0, 1, 0, &transform.matrix);
		VectorSet(&transform.trans, 0, 0, 0);
		transform.life_dec = 1;		//decrement life.
		tree->createEdge(nA, nB, transform);
	}

	tree->build();

	main_console->Print("created %d vtxs, %d tris\n",
		vertexBuffer.getCount(), triBuffer.getCount());

	overrideCameraKeys = 1;

	glEnable(GL_DEPTH_TEST);

	return true;
}

void Update_Application3D(void) {
	glDisable(GL_TEXTURE_2D);

	glCullFace(GL_NONE);

	glColor3f(1,1,1);
	glBegin(GL_TRIANGLES);
	for (int i = 0; i < triBuffer.getCount(); i++) {
		for (int j = 0; j < 3; j++) {
			glColor3fv(vertexBuffer.getVtxPtr(triBuffer.getTriVtxIndex(i,j)));
			glVertex3fv(vertexBuffer.getVtxPtr(triBuffer.getTriVtxIndex(i,j)));
		}
	}

	glEnd();
}

void Shutdown_Application3D(void) {
}
