/*

    File: Script.cpp
	Copyright (C) 2000-2007 Christopher Moore (christopher.e.moore@gmail.com)
	  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License along
    with this program; if not, write the Free Software Foundation, Inc., 51
    Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/
#include <math.h>

#include "../m_script/main.h"
#include "../m_foundation/main.h"	//m_cvar
#include "../m_foundation/random.h"

#include "scriptwnd.h"
#include "mesh.h"
#include "main.h"
#include "t3d.h"

CFUNC CFunc_runScript;
CFUNC CFunc_loadScript;

static CVarw *local_cvars[] = {
	new CVarw(0,				"run",			CFunc_runScript),
	new CVarw(0,				"load",			CFunc_loadScript),
};

// **************** BEGIN SCRIPT EXECUTOR ****************

//static const char *code = "";

static int PF_vertexRangeAlloc_params[] = {PT_INT};
static int PF_vertexRangeAlloc_return = PT_INT;
static void PF_vertexRangeAlloc(CThreadStackInterface *si) {
	int i;
	si->getParamInt(0, &i);
	int ret = vertexRangeAlloc(i);
	si->setReturnInt(ret);
}

static int PF_setRangeVtx_params[] = {PT_INT, PT_INT, PT_FLOAT, PT_FLOAT, PT_FLOAT, PT_FLOAT, PT_FLOAT};
static int PF_setRangeVtx_return = PT_VOID;
static void PF_setRangeVtx(CThreadStackInterface *si) {
	int i0, i1;
	float i2, i3, i4, i5, i6;
	si->getParamInt(0, &i0);
	si->getParamInt(1, &i1);
	si->getParamFloat(2, &i2);
	si->getParamFloat(3, &i3);
	si->getParamFloat(4, &i4);
	si->getParamFloat(5, &i5);
	si->getParamFloat(6, &i6);
	setRangeVtx(i0,i1,i2,i3,i4,i5,i6);
}

static int PF_BindRange_params[] = {PT_INT, PT_INT};
static int PF_BindRange_return = PT_VOID;
static void PF_BindRange(CThreadStackInterface *si) {
	int i0, i1;
	si->getParamInt(0, &i0);
	si->getParamInt(1, &i1);
	BindRange(i0,i1);
}

static int PF_T3DPush_return = PT_VOID;
static void PF_T3DPush(CThreadStackInterface *si) {
	T3DPush();
}

static int PF_T3DPop_return = PT_VOID;
static void PF_T3DPop(CThreadStackInterface *si) {
	T3DPop();
}

static int PF_T3DTranslate_params[] = {PT_FLOAT, PT_FLOAT, PT_FLOAT};
static int PF_T3DTranslate_return = PT_VOID;
static void PF_T3DTranslate(CThreadStackInterface *si) {
	float i0, i1, i2;
	si->getParamFloat(0, &i0);
	si->getParamFloat(1, &i1);
	si->getParamFloat(2, &i2);
	T3DTranslate(i0, i1, i2);
}

static int PF_T3DRotate_params[] = {PT_FLOAT, PT_FLOAT, PT_FLOAT, PT_FLOAT};
static int PF_T3DRotate_return = PT_VOID;
static void PF_T3DRotate(CThreadStackInterface *si) {
	float i0, i1, i2, i3;
	si->getParamFloat(0, &i0);
	si->getParamFloat(1, &i1);
	si->getParamFloat(2, &i2);
	si->getParamFloat(3, &i3);
	T3DRotate(i0, i1, i2, i3);
}

static int PF_T3DScale_params[] = {PT_FLOAT};
static int PF_T3DScale_return = PT_VOID;
static void PF_T3DScale(CThreadStackInterface *si) {
	float i0;
	si->getParamFloat(0, &i0);
	T3DScale(i0);
}

//static int PF_T3DSetScale_params[] = {PT_FLOAT};
//static int PF_T3DSetScale_return = PT_VOID;
//static void PF_T3DSetScale(CThreadStackInterface *si) {
//	float i0;
//	si->getParamFloat(0, &i0);
//	T3DSetScale(i0);
//}

static int PF_itof_params[] = {PT_INT};
static int PF_itof_return = PT_FLOAT;
static void PF_itof(CThreadStackInterface *si) {
	int i0;
	si->getParamInt(0, &i0);
	si->setReturnFloat((float)i0);
}

static int PF_MCRand_return = PT_FLOAT;
static void PF_MCRand(CThreadStackInterface *si) {
	si->setReturnFloat(MCRand());
}

static int PF_MSRand_params[] = {PT_INT};
static int PF_MSRand_return = PT_VOID;
static void PF_MSRand(CThreadStackInterface *si) {
	int seed;
	si->getParamInt(0, &seed);
	MSRand(seed);
}

static int PF_rrandf_params[] = {PT_FLOAT, PT_FLOAT};
static int PF_rrandf_return = PT_FLOAT;
static void PF_rrandf(CThreadStackInterface *si) {
	float min, max;
	si->getParamFloat(0, &min);
	si->getParamFloat(1, &max);
	si->setReturnFloat(min + MFRand() * (max - min));
}

static int PF_rrandi_params[] = {PT_INT, PT_INT};
static int PF_rrandi_return = PT_INT;
static void PF_rrandi(CThreadStackInterface *si) {
	int min, max;
	si->getParamInt(0, &min);
	si->getParamInt(1, &max);
	int ret = min;
	if (max > min) ret += MRand() % (max - min + 1);
	si->setReturnInt(ret);
}

static int PF_setTexture_params[] = {PT_STRING};
static int PF_setTexture_return = PT_VOID;
static void PF_setTexture(CThreadStackInterface *si) {
	int i0;
	si->getParamString(0, &i0);
	if (i0 == -1) return;
	meshSetTexture( si->program->stringTable[i0].c_str() );
}

static int PF_cos_params[] = {PT_FLOAT};
static int PF_cos_return = PT_FLOAT;
static void PF_cos(CThreadStackInterface *si) {
	float f0;
	si->getParamFloat(0, &f0);
	si->setReturnFloat( (float)cos(f0));
}

static int PF_sin_params[] = {PT_FLOAT};
static int PF_sin_return = PT_FLOAT;
static void PF_sin(CThreadStackInterface *si) {
	float f0;
	si->getParamFloat(0, &f0);
	si->setReturnFloat( (float)sin(f0));
}

void buildPredefined(CProgram *pr) {

	//vertexRangeAlloc
	pr->registerPredefFunction(
		"vertexRangeAlloc",
		pr->registerFuncType(CFuncType(
			PF_vertexRangeAlloc_return,
			numberof(PF_vertexRangeAlloc_params),
			PF_vertexRangeAlloc_params)),
			PF_vertexRangeAlloc);

	//setRangeVtx
	pr->registerPredefFunction(
		"setRangeVtx",
		pr->registerFuncType(CFuncType(
			PF_setRangeVtx_return,
			numberof(PF_setRangeVtx_params),
			PF_setRangeVtx_params)),
			PF_setRangeVtx);

	//BindRange
	pr->registerPredefFunction(
		"BindRange",
		pr->registerFuncType(CFuncType(
			PF_BindRange_return,
			numberof(PF_BindRange_params),
			PF_BindRange_params)),
			PF_BindRange);

	//T3DPush
	pr->registerPredefFunction(
		"T3DPush",
		pr->registerFuncType(CFuncType(
			PF_T3DPush_return,
			0,
			NULL)),
			PF_T3DPush);

	//T3DPop
	pr->registerPredefFunction(
		"T3DPop",
		pr->registerFuncType(CFuncType(
			PF_T3DPop_return,
			0,
			NULL)),
			PF_T3DPop);

	//T3DTranslate
	pr->registerPredefFunction(
		"T3DTranslate",
		pr->registerFuncType(CFuncType(
			PF_T3DTranslate_return,
			numberof(PF_T3DTranslate_params),
			PF_T3DTranslate_params)),
			PF_T3DTranslate);

	//T3DRotate
	pr->registerPredefFunction(
		"T3DRotate",
		pr->registerFuncType(CFuncType(
			PF_T3DRotate_return,
			numberof(PF_T3DRotate_params),
			PF_T3DRotate_params)),
			PF_T3DRotate);

	//T3DScale
	pr->registerPredefFunction(
		"T3DScale",
		pr->registerFuncType(CFuncType(
			PF_T3DScale_return,
			numberof(PF_T3DScale_params),
			PF_T3DScale_params)),
			PF_T3DScale);

//	//T3DSetScale
//	pr->registerPredefFunction(
//		"T3DSetScale",
//		pr->registerFuncType(CFuncType(
//			PF_T3DSetScale_return,
//			numberof(PF_T3DSetScale_params),
//			PF_T3DSetScale_params)),
//			PF_T3DSetScale);

	//itof
	pr->registerPredefFunction(
		"itof",
		pr->registerFuncType(CFuncType(
			PF_itof_return,
			numberof(PF_itof_params),
			PF_itof_params)),
			PF_itof);

	//MSRand
	pr->registerPredefFunction(
		"MSRand",
		pr->registerFuncType(CFuncType(
			PF_MSRand_return,
			numberof(PF_MSRand_params),
			PF_MSRand_params)),
			PF_MSRand);

	//MCRand
	pr->registerPredefFunction(
		"MCRand",
		pr->registerFuncType(CFuncType(
			PF_MCRand_return,
			0,
			NULL)),
			PF_MCRand);

	//rrandf
	pr->registerPredefFunction(
		"rrandf",
		pr->registerFuncType(CFuncType(
			PF_rrandf_return,
			numberof(PF_rrandf_params),
			PF_rrandf_params)),
			PF_rrandf);

	//rrandi
	pr->registerPredefFunction(
		"rrandi",
		pr->registerFuncType(CFuncType(
			PF_rrandi_return,
			numberof(PF_rrandi_params),
			PF_rrandi_params)),
			PF_rrandi);

	//cos
	pr->registerPredefFunction(
		"cos",
		pr->registerFuncType(CFuncType(
			PF_cos_return,
			numberof(PF_cos_params),
			PF_cos_params)),
			PF_cos);

	//sin
	pr->registerPredefFunction(
		"sin",
		pr->registerFuncType(CFuncType(
			PF_sin_return,
			numberof(PF_sin_params),
			PF_sin_params)),
			PF_sin);

	//setTexture
	pr->registerPredefFunction(
		"setTexture",
		pr->registerFuncType(CFuncType(
			PF_setTexture_return,
			numberof(PF_setTexture_params),
			PF_setTexture_params)),
			PF_setTexture);
}


void runScript(void) {
//d.g. init

	MSRand(0xdeadbeef);

	meshReset();
	T3DStackInit();

//script init
	int codeSize = 0;
	char *code = scriptWndGetCode(&codeSize);
	if (code && codeSize) {

		CLexemeParser lex(code, codeSize);
		CProgram program;
		buildPredefined(&program);
		CCodeCompiler compiler(&lex, &program);
		if (!compiler.compile()) {
			main_console->Print("error compiling: %s\n", compiler.getError());
		} else {
			//only do this post-compile
			CGlobalStack globalStack;
			
			if (!globalStack.reset(&program)) {
				main_console->Print("couldn't reset the global stack\n");
			} else {

				int mainID;
				if (!program.getFunctionID("main", &mainID)) {
					main_console->Print("couldn't find \"main\" function\n");
				} else {
					CScriptThread thread(&program, &globalStack);
					thread.runNamed("main");
				}
			}
		}
//		//out the goods
//		main_console->Print(" === begin program contents === \n");
//		program.print();
//		main_console->Print(" ==== end program contents ==== \n");

		free(code);
	}

	meshPostLoad();


//	for (int i = 0; i < vertexBuffer.getCount(); i++) {
//		const float *v = vertexBuffer.getVertexPtr(i);
//		main_console->Print("  %f %f %f\n", v[0], v[1], v[2]);
//	}
}

// ************************ script window interface *************************

void CFunc_loadScript(int argc, char **argv) {
	if (argc < 2) {
		main_console->Print("usage: %s <filename>\n", argv[0]);
		return;
	}

	scriptWndLoad(argv[1]);
}

void CFunc_runScript(int argc, char **argv) {
	if (argc > 1) {
		scriptWndLoad(argv[1]);
	}

	runScript();
}
