/*

    File: ScriptMenu.cpp
	Copyright (C) 2000-2007 Christopher Moore (christopher.e.moore@gmail.com)
	  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License along
    with this program; if not, write the Free Software Foundation, Inc., 51
    Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/
#include <windows.h>
#include <stdio.h>
#include <direct.h>

#include "../m_system3d/main.h"
#include "../m_system3d/error.h"
#include "../m_foundation/main.h"

#include "scriptmenu.h"
#include "scriptwnd.h"
#include "script.h"
#include "resource.h"
#include "main.h"

//**************** menu IO

#define SCRIPT_EXT			".script"
#define SCRIPT_FILETYPE		"*.script\0*.script\0\0",	"script"

static char editor_filename[256] = {0};
static char editor_full_filename[256] = {0};

static BOOL EditorFunc_OpenFileDialog(const char *filter, const char *ext) {

//prompt for save changes if we've recorded a modification?

	OPENFILENAME o = {
		sizeof(OPENFILENAME),				//length, in bytes, of structure
		main_display->displayWnd,			//handle to window owner
		GetModuleHandle(NULL),				//instance
		filter,								//filter text
		NULL,								//custom filter
		NULL,								//custom filter size
		1,									//default filter to use
		editor_full_filename,				//file name to recieve from open dialog
		sizeof(editor_full_filename),		//file name size
		editor_filename,					//file title (no pathname)
		sizeof(editor_filename),			//file title size
		NULL,								//initial directory
		NULL,								//dialog title
		OFN_PATHMUSTEXIST | OFN_FILEMUSTEXIST,//flags
		0,									//file name offset
		0,									//file extension offset
		ext,								//default file extension
		NULL,								//custom data
		NULL,								//hook
		NULL								//template name
	};

	return GetOpenFileName(&o);
}

static BOOL EditorFunc_SaveFileDialog(const char *filter, const char *ext) {
	OPENFILENAME fn = {
		sizeof(OPENFILENAME),			//lStructSize
		main_display->displayWnd,		//hwndOwner
		NULL,							//hInstance
		filter,							//lpstrFilter
		NULL,							//lpstrCustomFilter
		0,								//nMaxCustFilter
		1,								//nFilterIndex
		editor_full_filename,		//lpstrFile
		sizeof(editor_full_filename),//nMaxFile
		editor_filename,			//lpstrFileTitle
		sizeof(editor_filename),	//nMaxFileTitle
		NULL,							//lpstrInitialDir
		NULL,							//lpstrTitle
		OFN_PATHMUSTEXIST | OFN_OVERWRITEPROMPT,	//Flags
		0,								//nFileOffset
		0,								//nFileExtension
		ext,							//lpstrDefExt
		NULL,							//lCustData
		NULL,							//lpfnHook
		NULL							//lpTemplateName
	};

	return GetSaveFileName(&fn);
}


static bool ScriptMenu_CommandCallback(WORD wNotifyCode, WORD wID, HWND hwndCtl) {

	if (wNotifyCode) return false;	//wNotifyCode == 0 when using menu input

	switch (wID) {
	case ID_FILE_NEW_WORLD:
		scriptWndSetCode("");
		runScript();
		return true;

	case ID_FILE_OPEN_WORLD:
		{
			char cwd[256];
			_getcwd(cwd, sizeof(cwd));
			BOOL success = EditorFunc_OpenFileDialog(SCRIPT_FILETYPE);
			_chdir(cwd);
			if (success) {
				scriptWndLoad(editor_full_filename);
			}
		}
		return true;

	case ID_FILE_SAVE_WORLD:
		{
			char cwd[256];
			_getcwd(cwd, sizeof(cwd));
			BOOL success = EditorFunc_SaveFileDialog(SCRIPT_FILETYPE);
			_chdir(cwd);
			if (success) {
				if (!scriptWndSave(editor_full_filename)) {
					Error("unable to save to file %s", editor_full_filename);
				}
			}
		}
		return true;

	case ID_FILE_EXIT_WORLD:
		RequestExit();
		return true;

	case ID_WINDOW_SCRIPT:
		scriptWndShow(2);
		return true;
	}

	return false;
}

static void Editor_DropFiles_Test(HDROP hDrop, const char *ext, void (*func)(const char *fn)) {
	char filename[256];
	int fileCount = DragQueryFile(hDrop, 0xFFFFFFFF, 0, 0);

	for (int i = 0; i < fileCount; i++) {

		memset(filename, 0, sizeof(filename));
		DragQueryFile(hDrop, i, filename, sizeof(filename));

		int namelen = strlen(filename);

		for (int j = namelen - 1; j >= 0; j--) {
			if (filename[j] == '.') break;
			if (filename[j] == '\\' || filename[j] == '/') break;
		}

		if (filename[j] != '.') continue;	//no extension on file

		if (!_stricmp(&filename[j], ext)) {
			func(filename);
		}
	}
}

static void scriptMenu_DoOpen(const char *filename) {
	if (!scriptWndLoad(filename)) {
		main_console->Print("unable to load to file %s", filename);
	}
}

static bool ScriptMenu_MessageCallback(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam) {
	switch (uMsg) {
	case WM_DROPFILES:
		Editor_DropFiles_Test((HDROP)wParam, SCRIPT_EXT, scriptMenu_DoOpen);
		DragFinish((HDROP)wParam);
		return true;
	}

	return false;
}

HMENU scriptMenu = NULL;

bool scriptMenuInit(void) {
	scriptMenu = LoadMenu(GetModuleHandle(NULL), MAKEINTRESOURCE(IDR_MENU_MAIN));
	if (!scriptMenu) {
		main_console->Print("failed to load the menu\n");
		return false;
	}

	main_display->commandCallback = ScriptMenu_CommandCallback;
	main_display->messageCallback = ScriptMenu_MessageCallback;
	DragAcceptFiles(main_display->displayWnd, TRUE);

	if (!SetMenu(main_display->displayWnd, scriptMenu)) {
		main_console->Print("failed to set the scriptMenu to the window\n");
		return false;
	}

	return true;
}

void scriptMenuKill(void) {
	if (scriptMenu) {
		DestroyMenu(scriptMenu);
		scriptMenu = NULL;
	}

	SetMenu(main_display->displayWnd, NULL);

//TODO - what happens when some other code module wants to take hold of these features too?
//interrupt chains, etc . . .
//	main_display->commandCallback = NULL;
//	main_display->messageCallback = NULL;
//	DragAcceptFiles(main_display->displayWnd, FALSE);
}
