/*

    File: ScriptWnd.cpp
	Copyright (C) 2000-2007 Christopher Moore (christopher.e.moore@gmail.com)
	  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License along
    with this program; if not, write the Free Software Foundation, Inc., 51
    Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/
#include <windows.h>	//windows functions
#include <stdio.h>		//file io

#include "../m_system3d/main.h"		//main_display
#include "../m_foundation/main.h"	//main_console

#include "scriptwnd.h"
#include "resource.h"
#include "main.h"

CFUNC CFunc_showScript;

static CVarw *local_cvars[] = {
	new CVarw(0,		"script",		CFunc_showScript),
};

HWND scriptWnd = NULL;

void scriptWndShow(int show) {
	if (!scriptWnd) return;

	LONG flags = GetWindowLong(scriptWnd, GWL_STYLE);
	if (show < 0 || show > 1) {
		show = (flags & WS_VISIBLE) ? 0 : 1;
	}
	ShowWindow(scriptWnd, show ? SW_SHOW : SW_HIDE);
}

void CFunc_showScript(int argc, char **argv) {
	scriptWndShow(1);
}

char *scriptWndGetCode(int *size) {
	HWND hWndEdit = GetDlgItem(scriptWnd, IDC_EDIT_SCRIPT);
	*size = GetWindowTextLength(hWndEdit) + 1;
	char *buffer = (char *)malloc(*size);
	if (buffer) {
		GetWindowText(hWndEdit, buffer, *size);
	}
	return buffer;
}

void scriptWndSetCode(const char *code) {
	SetWindowText(GetDlgItem(scriptWnd, IDC_EDIT_SCRIPT), code);
}

bool scriptWndSave(const char *filename) {
	FILE *file = fopen(filename, "wb");
	if (!file) {
		main_console->Print("scriptWndSave - fopen failed\n");
		return false;
	}

	int size;
	char *buffer;
	if (!(buffer = scriptWndGetCode(&size))) {
		main_console->Print("scriptWndSave - malloc failed\n");
		fclose(file);
		return false;
	}


	if (fwrite(buffer, size, 1, file) != 1) {
		main_console->Print("scriptWndSave - fwrite failed\n");
		fclose(file);
		free(buffer);
		return false;
	}

	fclose(file);

	free(buffer);

	return true;
}

bool scriptWndLoad(const char *filename) {
	FILE *file = fopen(filename, "rb");

	if (!scriptWnd) return false;

	if (!file) {
		main_console->Print("loadScript - fopen failed\n");
		return false;
	}

	if (fseek(file, 0, SEEK_END)) {
		main_console->Print("loadScript - fseek failed\n");
		fclose(file);
		return false;
	}

	int size = ftell(file);
	if (size == -1) {
		main_console->Print("loadScript - ftell failed\n");
		fclose(file);
		return false;
	}

	char *buffer = (char *)malloc(size+1);
	if (!buffer) {
		main_console->Print("loadScript - malloc failed\n");
		fclose(file);
		return false;
	}
	buffer[size]=0;

	if (fseek(file, 0, SEEK_SET)) {
		main_console->Print("loadScript - fseek failed\n");
		fclose(file);
		free(buffer);
		return false;
	}

	if (fread(buffer, size, 1, file) != 1) {
		main_console->Print("loadScript - fread failed\n");
		fclose(file);
		free(buffer);
		return false;
	}

	fclose(file);

//	main_console->Print("Script: %s\n", buffer);

	scriptWndSetCode(buffer);

	free(buffer);

	return true;
}

//font handle for the script editor
static HFONT hEditScriptFont = 0;

//script window callback method
static BOOL CALLBACK ScriptWndDialogProc( HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam) {
	switch (uMsg) {
	case WM_INITDIALOG:

		//because it hasnt figured it out yet
		scriptWnd = hwndDlg;

		scriptWndSetCode("//code goes here\r\n");

		SendMessage(GetDlgItem(scriptWnd, IDC_EDIT_SCRIPT), WM_SETFONT, (WPARAM)hEditScriptFont, MAKELPARAM(false, 0));

		return false;

	case WM_CLOSE:
		scriptWndShow(false);
		return true;

	case WM_SIZE:
		SetWindowPos(
			GetDlgItem(scriptWnd, IDC_EDIT_SCRIPT),
			NULL,
			0,
			0,
			LOWORD(lParam),  // width of client area
			HIWORD(lParam), // height of client area 
			SWP_DRAWFRAME | SWP_NOZORDER | SWP_NOOWNERZORDER | SWP_NOMOVE );
		return true;
	}
	return false;
}

bool scriptWndInit(void) {

	// *********** font init

	HDC hDC = GetDC(main_display->displayWnd);

	//special case, for the script editor
	hEditScriptFont = CreateFont(
		-MulDiv(10, GetDeviceCaps(hDC, LOGPIXELSY), 72),  // logical height of font 
		0,  // logical average character width 
		0,  // angle of escapement 
		0,  // base-line orientation angle 
		FW_REGULAR,  // font weight 
		0,  // italic attribute flag 
		0,  // underline attribute flag 
		0,  // strikeout attribute flag 
		DEFAULT_CHARSET,  // character set identifier 
		OUT_TT_PRECIS,  // output precision 
		CLIP_DEFAULT_PRECIS,  // clipping precision 
		DEFAULT_QUALITY,  // output quality 
		DEFAULT_PITCH | FF_MODERN,  // pitch and family 
		"Courier New"		 // pointer to typeface name string 
	);

	ReleaseDC(main_display->displayWnd, hDC);

	// ********* window init

	scriptWnd = CreateDialog(
		GetModuleHandle(NULL),
		MAKEINTRESOURCE(IDD_DIALOG_SCRIPT),
		main_display->displayWnd,
		(DLGPROC)ScriptWndDialogProc);

	if (!scriptWnd) {
		main_console->Print("failed to create the window\n");
	}

	return !!scriptWnd && !!hEditScriptFont;
}

void scriptWndKill(void) {
	if (scriptWnd) {
		DestroyWindow(scriptWnd);
		scriptWnd = NULL;
	}
	if (hEditScriptFont) {
		DeleteObject(hEditScriptFont);
		hEditScriptFont = NULL;
	}
}